"""CLI commands which use the application framework"""

import asyncio
import binascii

import click

import bellows.types as t
import bellows.ezsp
import bellows.zigbee.application
import zigpy.endpoint
from . import opts
from . import util
from .main import main


@main.command()
@opts.database_file
@opts.channel
@opts.extended_pan
@opts.pan
@click.pass_context
def form(ctx, database, channel, pan_id, extended_pan_id):
    """Form a new ZigBee network"""
    ctx.obj['database_file'] = database

    async def inner(ctx):
        app = ctx.obj['app']
        await app.initialize()
        await app.form_network(channel, pan_id, extended_pan_id)

    return util.app(inner, app_startup=False)(ctx)


@main.command()
@opts.database_file
@opts.duration_s
@click.pass_context
def permit(ctx, database, duration_s):
    """Allow devices to join this ZigBee network"""
    ctx.obj['database_file'] = database

    async def inner(ctx):
        app = ctx.obj['app']
        await app.permit(duration_s)

        click.echo("Joins are permitted for the next %ss..." % (duration_s, ))
        await asyncio.sleep(duration_s + 1)
        click.echo("Done")

    return util.app(inner)(ctx)


@main.command()
@click.argument('node', type=util.ZigbeeNodeParamType())
@click.argument('code')
@opts.database_file
@opts.duration_s
@click.pass_context
def permit_with_key(ctx, database, duration_s, node, code):
    """Allow devices to join this ZigBee network using an install code"""
    ctx.obj['database_file'] = database
    code = binascii.unhexlify(code)

    async def inner(ctx):
        app = ctx.obj['app']
        try:
            await app.permit_with_key(node, code, duration_s)

            click.echo("Joins are permitted for the next %ss..." % (duration_s, ))
            await asyncio.sleep(duration_s + 1)
            click.echo("Done")
        except Exception as e:
            click.echo(e)

    return util.app(inner)(ctx)


@main.command()
@opts.database_file
@click.pass_context
def devices(ctx, database):
    """Show device database"""

    def print_clusters(title, clusters):
        clusters = sorted(list(clusters.items()))
        if clusters:
            click.echo("      %s:" % (title, ))
        for cluster_id, cluster in clusters:
            click.echo("        %s (%s)" % (
                cluster.name, cluster_id
            ))

    ezsp = bellows.ezsp.EZSP()
    app = bellows.zigbee.application.ControllerApplication(ezsp, database)
    for ieee, dev in app.devices.items():
        click.echo("Device:")
        click.echo("  NWK: 0x%04x" % (dev.nwk, ))
        click.echo("  IEEE: %s" % (ieee, ))
        click.echo("  Endpoints:")
        for epid, ep in dev.endpoints.items():
            if epid == 0:
                continue
            if ep.status == zigpy.endpoint.Status.NEW:
                click.echo("    %s: Uninitialized")
            else:
                click.echo(
                    "    %s: profile=0x%02x, device_type=%s" % (
                        epid,
                        ep.profile_id,
                        ep.device_type,
                    )
                )
                print_clusters("Input Clusters", ep.in_clusters)
                print_clusters("Output Clusters", ep.out_clusters)


@main.group()
@click.pass_context
@opts.database_file
@click.argument('node', type=util.ZigbeeNodeParamType())
def zdo(ctx, node, database):
    """Perform ZDO operations against a device"""
    ctx.obj['node'] = node
    ctx.obj['database_file'] = database


@zdo.command()
@click.pass_context
@util.app
async def endpoints(ctx):
    """List endpoints on a device"""
    app = ctx.obj['app']
    node = ctx.obj['node']

    dev = util.get_device(app, node)
    if dev is None:
        return

    try:
        v = await dev.zdo.request(0x0005, dev.nwk)
        if v[0] != t.EmberStatus.SUCCESS:
            click.echo("Non-success response: %s" % (v, ))
        else:
            click.echo(v[2])
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@util.app
@click.argument('endpoint', type=click.IntRange(1, 255))
async def get_endpoint(ctx, endpoint):
    """Show an endpoint's simple descriptor"""
    app = ctx.obj['app']
    node = ctx.obj['node']

    dev, endp = util.get_endpoint(app, node, endpoint)
    if endp is None:
        return

    try:
        v = await dev.zdo.request(0x0004, dev.nwk, endpoint)
        if v[0] != t.EmberStatus.SUCCESS:
            click.echo("Non-success response: %s" % (v, ))
        else:
            click.echo(v[2])
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@click.argument('endpoint', type=click.IntRange(1, 255))
@click.argument('cluster', type=click.IntRange(0, 65535))
@util.app
async def bind(ctx, endpoint, cluster):
    """Bind to a cluster on a node"""
    app = ctx.obj['app']
    node = ctx.obj['node']

    dev, endp, clust = util.get_in_cluster(app, node, endpoint, cluster)
    if clust is None:
        return

    try:
        v = await dev.zdo.bind(endpoint, cluster)
        click.echo(v)
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@click.argument('endpoint', type=click.IntRange(1, 255))
@click.argument('cluster', type=click.IntRange(0, 65535))
@util.app
async def unbind(ctx, endpoint, cluster):
    """Unbind a cluster on a node"""
    app = ctx.obj['app']
    node = ctx.obj['node']

    dev, endp, clust = util.get_in_cluster(app, node, endpoint, cluster)
    if clust is None:
        return

    try:
        v = await dev.zdo.unbind(endpoint, cluster)
        click.echo(v)
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@util.app
async def leave(ctx):
    """Tell a node to leave the network"""
    app = ctx.obj['app']

    try:
        v = await app.remove(ctx.obj['node'])
        click.echo(v)
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@main.group()
@click.pass_context
@opts.database_file
@click.argument('node', type=util.ZigbeeNodeParamType())
@click.argument('endpoint', type=click.IntRange(1, 255))
@click.argument('cluster', type=click.IntRange(0, 65535))
def zcl(ctx, database, node, cluster, endpoint):
    """Peform ZCL operations against a device"""
    ctx.obj['database_file'] = database
    ctx.obj['node'] = node
    ctx.obj['endpoint'] = endpoint
    ctx.obj['cluster'] = cluster


@zcl.command()
@click.pass_context
@click.argument('attribute', type=click.IntRange(0, 65535))
@opts.manufacturer
@util.app
async def read_attribute(ctx, attribute, manufacturer):
    app = ctx.obj['app']
    node = ctx.obj['node']
    endpoint_id = ctx.obj['endpoint']
    cluster_id = ctx.obj['cluster']

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await cluster.read_attributes([attribute], allow_cache=False, manufacturer=manufacturer)
        if not v:
            click.echo("Received empty response")
        elif attribute not in v[0]:
            click.echo("Attribute %s not successful. Status=%s" % (
                attribute, v[1][attribute]
            ))
        else:
            click.echo("%s=%s" % (attribute, v[0][attribute]))
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@zcl.command()
@click.pass_context
@click.argument('attribute', type=click.IntRange(0, 65535))
@click.argument('value', type=click.IntRange(0, 65535))
@opts.manufacturer
@util.app
async def write_attribute(ctx, attribute, value, manufacturer):
    app = ctx.obj['app']
    node = ctx.obj['node']
    endpoint_id = ctx.obj['endpoint']
    cluster_id = ctx.obj['cluster']

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await cluster.write_attributes({attribute: value}, manufacturer=manufacturer)
        click.echo(v)
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@zcl.command()
@click.pass_context
def commands(ctx):
    database = ctx.obj['database_file']
    node = ctx.obj['node']
    endpoint_id = ctx.obj['endpoint']
    cluster_id = ctx.obj['cluster']

    ezsp = bellows.ezsp.EZSP()
    app = bellows.zigbee.application.ControllerApplication(ezsp, database)

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    for c in cluster.commands:
        click.echo(c)


@zcl.command()
@click.pass_context
@click.argument('command')
@click.argument('parameters', nargs=-1)
@opts.manufacturer
@util.app
async def command(ctx, command, parameters, manufacturer):
    app = ctx.obj['app']
    node = ctx.obj['node']
    endpoint_id = ctx.obj['endpoint']
    cluster_id = ctx.obj['cluster']

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await getattr(cluster, command)(*parameters, manufacturer=manufacturer)
        click.echo(v)
    except ValueError as e:
        click.echo(e)
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)


@zcl.command()
@click.pass_context
@click.argument('attribute', type=click.IntRange(0, 65535))
@click.argument('min_interval', type=click.IntRange(0, 65535))
@click.argument('max_interval', type=click.IntRange(0, 65535))
@click.argument('reportable_change', type=click.INT)
@util.app
async def configure_reporting(ctx,
                              attribute,
                              min_interval,
                              max_interval,
                              reportable_change):
    app = ctx.obj['app']
    node = ctx.obj['node']
    endpoint_id = ctx.obj['endpoint']
    cluster_id = ctx.obj['cluster']

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await cluster.configure_reporting(
            attribute,
            min_interval,
            max_interval,
            reportable_change,
        )
        click.echo(v)
    except bellows.zigbee.exceptions.DeliveryError as e:
        click.echo(e)



import asyncio
import binascii
import functools
import logging

import bellows.types as t
import bellows.uart as uart
from bellows.commands import COMMANDS
from bellows.exception import EzspError

EZSP_CMD_TIMEOUT = 10
LOGGER = logging.getLogger(__name__)


class EZSP:

    COMMANDS = COMMANDS
    ezsp_version = 4

    def __init__(self):
        self._awaiting = {}
        self._baudrate = None
        self._callbacks = {}
        self._device = None
        self._ezsp_event = asyncio.Event()
        self._seq = 0
        self._gw = None
        self._awaiting = {}
        self.COMMANDS_BY_ID = {}
        for name, details in self.COMMANDS.items():
            self.COMMANDS_BY_ID[details[0]] = (name, details[1], details[2])

    async def connect(self, device, baudrate):
        assert self._gw is None
        self._baudrate = baudrate
        self._device = device
        self._gw = await uart.connect(device, baudrate, self)

    def reconnect(self):
        """Reconnect using saved parameters."""
        LOGGER.debug("Reconnecting %s serial port on %s bauds",
                     self._device, self._baudrate)
        return self.connect(self._device, self._baudrate)

    async def reset(self):
        LOGGER.debug("Resetting EZSP")
        self.stop_ezsp()
        for seq in self._awaiting:
            future = self._awaiting[seq][2]
            if not future.done():
                future.cancel()
        self._awaiting = {}
        self._callbacks = {}
        self._seq = 0
        await self._gw.reset()
        self.start_ezsp()

    async def version(self):
        version = self.ezsp_version
        result = await self._command('version', version)
        if result[0] != version:
            LOGGER.debug("Switching to eszp version %d", result[0])
            await self._command('version', result[0])

    def close(self):
        self.stop_ezsp()
        if self._gw:
            self._gw.close()
            self._gw = None

    def _ezsp_frame(self, name, *args):
        c = self.COMMANDS[name]
        data = t.serialize(args, c[1])
        frame = [
            self._seq & 0xff,
            0,    # Frame control. TODO.
            c[0]  # Frame ID
        ]
        if self.ezsp_version >= 5:
            frame.insert(1, 0xFF)  # Legacy Frame ID
            frame.insert(1, 0x00)  # Ext frame control. TODO.

        return bytes(frame) + data

    def _command(self, name, *args):
        LOGGER.debug("Send command %s: %s", name, args)
        if not self.is_ezsp_running:
            raise EzspError("EZSP is not running")

        data = self._ezsp_frame(name, *args)
        self._gw.data(data)
        c = self.COMMANDS[name]
        future = asyncio.Future()
        self._awaiting[self._seq] = (c[0], c[2], future)
        self._seq = (self._seq + 1) % 256
        return asyncio.wait_for(future, timeout=EZSP_CMD_TIMEOUT)

    async def _list_command(self, name, item_frames, completion_frame, spos, *args):
        """Run a command, returning result callbacks as a list"""
        fut = asyncio.Future()
        results = []

        def cb(frame_name, response):
            if frame_name in item_frames:
                results.append(response)
            elif frame_name == completion_frame:
                fut.set_result(response)

        cbid = self.add_callback(cb)
        try:
            v = await self._command(name, *args)
            if v[0] != t.EmberStatus.SUCCESS:
                raise Exception(v)
            v = await fut
            if v[spos] != t.EmberStatus.SUCCESS:
                raise Exception(v)
        finally:
            self.remove_callback(cbid)

        return results

    startScan = functools.partialmethod(
        _list_command,
        'startScan',
        ['energyScanResultHandler', 'networkFoundHandler'],
        'scanCompleteHandler',
        1,
    )
    pollForData = functools.partialmethod(
        _list_command,
        'pollForData',
        ['pollHandler'],
        'pollCompleteHandler',
        0,
    )
    zllStartScan = functools.partialmethod(
        _list_command,
        'zllStartScan',
        ['zllNetworkFoundHandler'],
        'zllScanCompleteHandler',
        0,
    )
    rf4ceDiscovery = functools.partialmethod(
        _list_command,
        'rf4ceDiscovery',
        ['rf4ceDiscoveryResponseHandler'],
        'rf4ceDiscoveryCompleteHandler',
        0,
    )

    def connection_lost(self, exc):
        """Lost serial connection."""
        LOGGER.debug("%s connection lost unexpectedly: %s", self._device, exc)
        self.enter_failed_state("Serial connection loss: {}".format(exc))

    def enter_failed_state(self, error):
        """UART received error frame."""
        LOGGER.error(
            "NCP entered failed state. Requesting APP controller restart")
        self.stop_ezsp()
        self.handle_callback('_reset_controller_application', (error, ))

    async def formNetwork(self, parameters):  # noqa: N802
        fut = asyncio.Future()

        def cb(frame_name, response):
            nonlocal fut
            if frame_name == 'stackStatusHandler':
                fut.set_result(response)

        self.add_callback(cb)
        v = await self._command('formNetwork', parameters)
        if v[0] != t.EmberStatus.SUCCESS:
            raise Exception("Failure forming network: %s" % (v, ))

        v = await fut
        if v[0] != t.EmberStatus.NETWORK_UP:
            raise Exception("Failure forming network: %s" % (v, ))

        return v

    def __getattr__(self, name):
        if name not in self.COMMANDS:
            raise AttributeError("%s not found in COMMANDS" % (name, ))

        return functools.partial(self._command, name)

    def frame_received(self, data):
        """Handle a received EZSP frame

        The protocol has taken care of UART specific framing etc, so we should
        just have EZSP application stuff here, with all escaping/stuffing and
        data randomization removed.
        """
        sequence, frame_id, data = data[0], data[2], data[3:]
        if frame_id == 0xFF:
            frame_id = 0
            if len(data) > 1:
                frame_id = data[1]
                data = data[2:]

        frame_name = self.COMMANDS_BY_ID[frame_id][0]
        LOGGER.debug(
            "Application frame %s (%s) received: %s",
            frame_id,
            frame_name, binascii.hexlify(data)
        )

        if sequence in self._awaiting:
            expected_id, schema, future = self._awaiting.pop(sequence)
            assert expected_id == frame_id
            result, data = t.deserialize(data, schema)
            try:
                future.set_result(result)
            except asyncio.InvalidStateError as exc:
                LOGGER.debug(
                    "Error processing %s response. %s command timed out?",
                    sequence, self.COMMANDS_BY_ID.get(expected_id,
                                                      [expected_id])[0]
                )
        else:
            schema = self.COMMANDS_BY_ID[frame_id][2]
            frame_name = self.COMMANDS_BY_ID[frame_id][0]
            result, data = t.deserialize(data, schema)
            self.handle_callback(frame_name, result)

        if frame_id == 0x00:
            self.ezsp_version = result[0]

    def add_callback(self, cb):
        id_ = hash(cb)
        while id_ in self._callbacks:
            id_ += 1
        self._callbacks[id_] = cb
        return id_

    def remove_callback(self, id_):
        return self._callbacks.pop(id_)

    def handle_callback(self, *args):
        for callback_id, handler in self._callbacks.items():
            try:
                handler(*args)
            except Exception as e:
                LOGGER.exception("Exception running handler", exc_info=e)

    def start_ezsp(self):
        """Mark EZSP as running."""
        self._ezsp_event.set()

    def stop_ezsp(self):
        """Mark EZSP stopped."""
        self._ezsp_event.clear()

    @property
    def is_ezsp_running(self):
        """Return True if EZSP is running."""
        return self._ezsp_event.is_set()



import asyncio
import binascii
import logging
import os

from serial import SerialException
from zigpy.exceptions import DeliveryError
from zigpy.quirks import CustomDevice, CustomEndpoint
from zigpy.types import BroadcastAddress
import zigpy.application
import zigpy.device
import zigpy.util
import zigpy.zdo

import bellows.types as t
import bellows.zigbee.util
from bellows.exception import ControllerError, EzspError
import bellows.multicast

APS_ACK_TIMEOUT = 120
APS_REPLY_TIMEOUT = 5
APS_REPLY_TIMEOUT_EXTENDED = 28
MAX_WATCHDOG_FAILURES = 4
RESET_ATTEMPT_BACKOFF_TIME = 5
WATCHDOG_WAKE_PERIOD = 10

LOGGER = logging.getLogger(__name__)


class ControllerApplication(zigpy.application.ControllerApplication):
    direct = t.EmberOutgoingMessageType.OUTGOING_DIRECT

    def __init__(self, ezsp, database_file=None):
        super().__init__(database_file=database_file)
        self._ctrl_event = asyncio.Event()
        self._ezsp = ezsp
        self._multicast = bellows.multicast.Multicast(ezsp)
        self._pending = Requests()
        self._watchdog_task = None
        self._reset_task = None
        self._in_flight_msg = None

    @property
    def controller_event(self):
        """Return asyncio.Event for controller app."""
        return self._ctrl_event

    @property
    def is_controller_running(self):
        """Return True if controller was successfully initialized."""
        return self.controller_event.is_set() and self._ezsp.is_ezsp_running

    @property
    def multicast(self):
        """Return EZSP MulticastController."""
        return self._multicast

    async def initialize(self):
        """Perform basic NCP initialization steps"""
        e = self._ezsp

        await e.reset()
        await e.version()

        c = t.EzspConfigId
        await self._cfg(c.CONFIG_STACK_PROFILE, 2)
        await self._cfg(c.CONFIG_SECURITY_LEVEL, 5)
        await self._cfg(c.CONFIG_SUPPORTED_NETWORKS, 1)
        zdo = (
            t.EmberZdoConfigurationFlags.APP_RECEIVES_SUPPORTED_ZDO_REQUESTS |
            t.EmberZdoConfigurationFlags.APP_HANDLES_UNSUPPORTED_ZDO_REQUESTS
        )
        await self._cfg(c.CONFIG_APPLICATION_ZDO_FLAGS, zdo)
        await self._cfg(c.CONFIG_TRUST_CENTER_ADDRESS_CACHE_SIZE, 2)
        await self._cfg(c.CONFIG_ADDRESS_TABLE_SIZE, 16)
        await self._cfg(c.CONFIG_SOURCE_ROUTE_TABLE_SIZE, 8)
        await self._cfg(c.CONFIG_MAX_END_DEVICE_CHILDREN, 32)
        await self._cfg(c.CONFIG_INDIRECT_TRANSMISSION_TIMEOUT, 7680)
        await self._cfg(c.CONFIG_KEY_TABLE_SIZE, 1)
        await self._cfg(c.CONFIG_TRANSIENT_KEY_TIMEOUT_S, 180, True)
        await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT, 60)
        await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT, 8)
        await self._cfg(c.CONFIG_MULTICAST_TABLE_SIZE,
                        self.multicast.TABLE_SIZE)
        await self._cfg(c.CONFIG_PACKET_BUFFER_COUNT, 0xff)

        status, count = await e.getConfigurationValue(
            c.CONFIG_APS_UNICAST_MESSAGE_COUNT)
        assert status == t.EmberStatus.SUCCESS
        self._in_flight_msg = asyncio.Semaphore(count)
        LOGGER.debug("APS_UNICAST_MESSAGE_COUNT is set to %s", count)

        await self.add_endpoint(
            output_clusters=[zigpy.zcl.clusters.security.IasZone.cluster_id]
        )

    async def add_endpoint(self, endpoint=1,
                           profile_id=zigpy.profiles.zha.PROFILE_ID,
                           device_id=0xbeef,
                           app_flags=0x00,
                           input_clusters=[],
                           output_clusters=[]):
        """Add endpoint."""
        res = await self._ezsp.addEndpoint(
            endpoint, profile_id, device_id, app_flags,
            len(input_clusters), len(output_clusters),
            input_clusters, output_clusters
        )
        LOGGER.debug("Ezsp adding endpoint: %s", res)

    async def startup(self, auto_form=False):
        """Perform a complete application startup"""
        await self.initialize()
        e = self._ezsp

        v = await e.networkInit()
        if v[0] != t.EmberStatus.SUCCESS:
            if not auto_form:
                raise Exception("Could not initialize network")
            await self.form_network()

        v = await e.getNetworkParameters()
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check
        if v[1] != t.EmberNodeType.COORDINATOR:
            if not auto_form:
                raise Exception("Network not configured as coordinator")

            LOGGER.info("Forming network")
            await self._ezsp.leaveNetwork()
            await asyncio.sleep(1)  # TODO
            await self.form_network()

        await self._policy()
        nwk = await e.getNodeId()
        self._nwk = nwk[0]
        ieee = await e.getEui64()
        self._ieee = ieee[0]

        e.add_callback(self.ezsp_callback_handler)
        self.controller_event.set()
        self._watchdog_task = asyncio.ensure_future(self._watchdog())

        self.handle_join(self.nwk, self.ieee, 0)
        LOGGER.debug("EZSP nwk=0x%04x, IEEE=%s", self._nwk, str(self._ieee))
        await self.multicast.startup(self.get_device(self.ieee))

    async def shutdown(self):
        """Shutdown and cleanup ControllerApplication."""
        LOGGER.info("Shutting down ControllerApplication")
        self.controller_event.clear()
        if self._watchdog_task and not self._watchdog_task.done():
            LOGGER.debug("Cancelling watchdog")
            self._watchdog_task.cancel()
        if self._reset_task and not self._reset_task.done():
            self._reset_task.cancel()
        self._ezsp.close()

    async def form_network(self, channel=15, pan_id=None, extended_pan_id=None):
        channel = t.uint8_t(channel)

        if pan_id is None:
            pan_id = t.uint16_t.from_bytes(os.urandom(2), 'little')
        pan_id = t.uint16_t(pan_id)

        if extended_pan_id is None:
            extended_pan_id = t.fixed_list(8, t.uint8_t)([t.uint8_t(0)] * 8)

        initial_security_state = bellows.zigbee.util.zha_security(controller=True)
        v = await self._ezsp.setInitialSecurityState(initial_security_state)
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check

        parameters = t.EmberNetworkParameters()
        parameters.panId = pan_id
        parameters.extendedPanId = extended_pan_id
        parameters.radioTxPower = t.uint8_t(8)
        parameters.radioChannel = channel
        parameters.joinMethod = t.EmberJoinMethod.USE_MAC_ASSOCIATION
        parameters.nwkManagerId = t.EmberNodeId(0)
        parameters.nwkUpdateId = t.uint8_t(0)
        parameters.channels = t.uint32_t(0)

        await self._ezsp.formNetwork(parameters)
        await self._ezsp.setValue(t.EzspValueId.VALUE_STACK_TOKEN_WRITING, 1)

    async def _cfg(self, config_id, value, optional=False):
        v = await self._ezsp.setConfigurationValue(config_id, value)
        if not optional:
            assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check

    async def _policy(self):
        """Set up the policies for what the NCP should do"""
        e = self._ezsp
        v = await e.setPolicy(
            t.EzspPolicyId.TC_KEY_REQUEST_POLICY,
            t.EzspDecisionId.DENY_TC_KEY_REQUESTS,
        )
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check
        v = await e.setPolicy(
            t.EzspPolicyId.APP_KEY_REQUEST_POLICY,
            t.EzspDecisionId.ALLOW_APP_KEY_REQUESTS,
        )
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check
        v = await e.setPolicy(
            t.EzspPolicyId.TRUST_CENTER_POLICY,
            t.EzspDecisionId.ALLOW_PRECONFIGURED_KEY_JOINS,
        )
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check

    async def force_remove(self, dev):
        # This should probably be delivered to the parent device instead
        # of the device itself.
        await self._ezsp.removeDevice(dev.nwk, dev.ieee, dev.ieee)

    def ezsp_callback_handler(self, frame_name, args):
        LOGGER.debug("Received %s frame with %s", frame_name, args)
        if frame_name == 'incomingMessageHandler':
            self._handle_frame(*args)
        elif frame_name == 'messageSentHandler':
            if args[4] != t.EmberStatus.SUCCESS:
                self._handle_frame_failure(*args)
            else:
                self._handle_frame_sent(*args)
        elif frame_name == 'trustCenterJoinHandler':
            if args[2] == t.EmberDeviceUpdate.DEVICE_LEFT:
                self.handle_leave(args[0], args[1])
            else:
                self.handle_join(args[0], args[1], args[4])
        elif frame_name == '_reset_controller_application':
            self._handle_reset_request(*args)

    def _handle_frame(self, message_type, aps_frame, lqi, rssi, sender, binding_index, address_index, message):
        try:
            device = self.get_device(nwk=sender)
        except KeyError:
            LOGGER.debug("No such device %s", sender)
            return

        device.radio_details(lqi, rssi)
        try:
            tsn, command_id, is_reply, args = self.deserialize(device, aps_frame.sourceEndpoint, aps_frame.clusterId, message)
        except ValueError as e:
            LOGGER.error("Failed to parse message (%s) on cluster %d, because %s", binascii.hexlify(message), aps_frame.clusterId, e)
            return

        if is_reply:
            self._handle_reply(device, aps_frame, tsn, command_id, args)
        else:
            self.handle_message(device, False, aps_frame.profileId, aps_frame.clusterId, aps_frame.sourceEndpoint, aps_frame.destinationEndpoint, tsn, command_id, args)

    @staticmethod
    def _dst_pp(addr, aps):
        """Return format string and args."""
        ep, cluster = aps.destinationEndpoint, aps.clusterId
        return '[0x%04x:%s:0x%04x]: ', (addr, ep, cluster)

    def _handle_reply(self, sender, aps_frame, tsn, command_id, args):
        try:
            request = self._pending[tsn]
            if request.reply:
                request.reply.set_result(args)
            return
        except KeyError:
            hdr, hdr_args = self._dst_pp(sender.nwk, aps_frame)
            LOGGER.debug(hdr + "Unexpected response TSN=%s command=%s args=%s",
                         *(hdr_args + (tsn, command_id, args)))
        except asyncio.futures.InvalidStateError as exc:
            hdr, hdr_args = self._dst_pp(sender.nwk, aps_frame)
            LOGGER.debug(hdr + "Invalid state on future - probably duplicate response: %s",
                         *(hdr_args + (exc, )))
            # We've already handled, don't drop through to device handler
            return

        self.handle_message(sender, True, aps_frame.profileId, aps_frame.clusterId, aps_frame.sourceEndpoint, aps_frame.destinationEndpoint, tsn, command_id, args)

    def _handle_frame_failure(self, message_type, destination, aps_frame, message_tag, status, message):
        hdr, hdr_args = self._dst_pp(destination, aps_frame)
        try:
            request = self._pending[message_tag]
            msg = hdr + "message send failure: %s"
            msg_args = (hdr_args + (status, ))
            request.send.set_exception(DeliveryError(msg % msg_args))
        except KeyError:
            LOGGER.debug(hdr + "Unexpected message send failure", *hdr_args)
        except asyncio.futures.InvalidStateError as exc:
            LOGGER.debug(hdr + "Invalid state on future - probably duplicate response: %s",
                         *(hdr_args + (exc, )))

    def _handle_frame_sent(self, message_type, destination, aps_frame, message_tag, status, message):
        try:
            request = self._pending[message_tag]
            request.send.set_result(True)
        except KeyError:
            hdr, hdr_args = self._dst_pp(destination, aps_frame)
            LOGGER.debug(hdr + "Unexpected message send notification", *hdr_args)
        except asyncio.futures.InvalidStateError as exc:
            hdr, hdr_args = self._dst_pp(destination, aps_frame)
            LOGGER.debug(hdr + "Invalid state on future - probably duplicate response: %s",
                         *(hdr_args + (exc, )))

    def _handle_reset_request(self, error):
        """Reinitialize application controller."""
        LOGGER.debug("Resetting ControllerApplication. Cause: '%s'", error)
        self.controller_event.clear()
        if self._reset_task:
            LOGGER.debug("Preempting ControllerApplication reset")
            self._reset_task.cancel()

        self._reset_task = asyncio.ensure_future(self._reset_controller_loop())

    async def _reset_controller_loop(self):
        """Keep trying to reset controller until we succeed."""
        self._watchdog_task.cancel()
        while True:
            try:
                await self._reset_controller()
                break
            except (asyncio.TimeoutError, SerialException) as exc:
                LOGGER.debug(
                    "ControllerApplication reset unsuccessful: %s", str(exc))
            await asyncio.sleep(RESET_ATTEMPT_BACKOFF_TIME)

        self._reset_task = None
        LOGGER.debug("ControllerApplication successfully reset")

    async def _reset_controller(self):
        """Reset Controller."""
        self._ezsp.close()
        await asyncio.sleep(0.5)
        await self._ezsp.reconnect()
        await self.startup()

    @zigpy.util.retryable_request
    async def request(self, nwk, profile, cluster, src_ep, dst_ep, sequence, data, expect_reply=True,
                      timeout=APS_REPLY_TIMEOUT):
        if not self.is_controller_running:
            raise ControllerError("ApplicationController is not running")

        aps_frame = t.EmberApsFrame()
        aps_frame.profileId = t.uint16_t(profile)
        aps_frame.clusterId = t.uint16_t(cluster)
        aps_frame.sourceEndpoint = t.uint8_t(src_ep)
        aps_frame.destinationEndpoint = t.uint8_t(dst_ep)
        aps_frame.options = t.EmberApsOption(
            t.EmberApsOption.APS_OPTION_RETRY |
            t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
        )
        aps_frame.groupId = t.uint16_t(0)
        aps_frame.sequence = t.uint8_t(sequence)

        try:
            dev = self.get_device(nwk=nwk)
            if expect_reply and dev.node_desc.is_end_device in (True, None):
                LOGGER.debug("Extending timeout for %s/0x%04x", dev.ieee, nwk)
                await self._ezsp.setExtendedTimeout(dev.ieee, True)
                timeout = APS_REPLY_TIMEOUT_EXTENDED
        except KeyError:
            pass
        with self._pending.new(sequence, expect_reply) as req:
            async with self._in_flight_msg:
                res = await self._ezsp.sendUnicast(self.direct, nwk, aps_frame,
                                                   sequence, data)
                if res[0] != t.EmberStatus.SUCCESS:
                    hdr, hdr_args = self._dst_pp(nwk, aps_frame)
                    msg = hdr + "message send failure: %s"
                    msg_args = (hdr_args + (res[0], ))
                    raise DeliveryError(msg % msg_args)

                res = await asyncio.wait_for(req.send, timeout=APS_ACK_TIMEOUT)

            if expect_reply:
                res = await asyncio.wait_for(req.reply, timeout)
        return res

    def permit_ncp(self, time_s=60):
        assert 0 <= time_s <= 254
        return self._ezsp.permitJoining(time_s)

    async def permit_with_key(self, node, code, time_s=60):
        if type(node) is not t.EmberEUI64:
            node = t.EmberEUI64([t.uint8_t(p) for p in node])

        key = zigpy.util.convert_install_code(code)
        if key is None:
            raise Exception("Invalid install code")

        v = await self._ezsp.addTransientLinkKey(node, key)
        if v[0] != t.EmberStatus.SUCCESS:
            raise Exception("Failed to set link key")

        v = await self._ezsp.setPolicy(
            t.EzspPolicyId.TC_KEY_REQUEST_POLICY,
            t.EzspDecisionId.GENERATE_NEW_TC_LINK_KEY,
        )
        if v[0] != t.EmberStatus.SUCCESS:
            raise Exception("Failed to change policy to allow generation of new trust center keys")

        return self._ezsp.permitJoining(time_s, True)

    async def broadcast(self, profile, cluster, src_ep, dst_ep, grpid, radius,
                        sequence, data,
                        broadcast_address=BroadcastAddress.RX_ON_WHEN_IDLE):
        if not self.is_controller_running:
            raise ControllerError("ApplicationController is not running")

        aps_frame = t.EmberApsFrame()
        aps_frame.profileId = t.uint16_t(profile)
        aps_frame.clusterId = t.uint16_t(cluster)
        aps_frame.sourceEndpoint = t.uint8_t(src_ep)
        aps_frame.destinationEndpoint = t.uint8_t(dst_ep)
        aps_frame.options = t.EmberApsOption(
            t.EmberApsOption.APS_OPTION_NONE
        )
        aps_frame.groupId = t.uint16_t(grpid)
        aps_frame.sequence = t.uint8_t(sequence)

        with self._pending.new(sequence) as req:
            async with self._in_flight_msg:
                res = await self._ezsp.sendBroadcast(broadcast_address,
                                                     aps_frame, radius,
                                                     sequence, data)
                if res[0] != t.EmberStatus.SUCCESS:
                    hdr, hdr_args = self._dst_pp(broadcast_address, aps_frame)
                    msg = hdr + "Broadcast failure: %s"
                    msg_args = (hdr_args + (res[0], ))
                    raise DeliveryError(msg % msg_args)

                # Wait for messageSentHandler message
                res = await asyncio.wait_for(req.send,
                                             timeout=APS_ACK_TIMEOUT)
        return res

    async def _watchdog(self):
        """Watchdog handler."""
        LOGGER.debug("Starting EZSP watchdog")
        failures = 0
        await asyncio.sleep(WATCHDOG_WAKE_PERIOD)
        while True:
            try:
                await asyncio.wait_for(self.controller_event.wait(),
                                       timeout=WATCHDOG_WAKE_PERIOD * 2)
                await self._ezsp.nop()
                failures = 0
            except (asyncio.TimeoutError, EzspError) as exc:
                LOGGER.warning("Watchdog heartbeat timeout: %s", str(exc))
                failures += 1
                if failures > MAX_WATCHDOG_FAILURES:
                    break
            await asyncio.sleep(WATCHDOG_WAKE_PERIOD)

        self._handle_reset_request(
            "Watchdog timeout. Heartbeat timeouts: {}".format(failures))


class Requests(dict):
    def new(self, sequence, expect_reply=False):
        """Wrap new request into a context manager."""
        return Request(self, sequence, expect_reply)


class Request:
    """Context manager."""

    def __init__(self, pending, sequence, expect_reply=False):
        """Init context manager for sendUnicast/sendBroadcast."""
        assert sequence not in pending
        self._pending = pending
        self._reply_fut = None
        if expect_reply:
            self._reply_fut = asyncio.Future()
        self._send_fut = asyncio.Future()
        self._sequence = sequence

    @property
    def reply(self):
        """Reply Future."""
        return self._reply_fut

    @property
    def sequence(self):
        """Send Future."""
        return self._sequence

    @property
    def send(self):
        return self._send_fut

    def __enter__(self):
        """Return context manager."""
        self._pending[self.sequence] = self
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Clean up pending on exit."""
        if not self.send.done():
            self.send.cancel()
        if self.reply and not self.reply.done():
            self.reply.cancel()
        self._pending.pop(self.sequence)

        return not exc_type


class EZSPCoordinator(CustomDevice):
    """Zigpy Device representing Coordinator."""

    class EZSPEndpoint(CustomEndpoint):
        async def add_to_group(self, grp_id: int,
                               name: str = None) -> t.EmberStatus:
            if grp_id in self.member_of:
                return t.EmberStatus.SUCCESS

            app = self.device.application
            status = await app.multicast.subscribe(grp_id)
            if status != t.EmberStatus.SUCCESS:
                self.debug("Couldn't subscribe to 0x%04x group", grp_id)
                return status

            group = app.groups.add_group(grp_id, name)
            group.add_member(self)
            return status

        async def remove_from_group(self, grp_id: int) -> t.EmberStatus:
            if grp_id not in self.member_of:
                return t.EmberStatus.SUCCESS

            app = self.device.application
            status = await app.multicast.unsubscribe(grp_id)
            if status != t.EmberStatus.SUCCESS:
                self.debug("Couldn't unsubscribe 0x%04x group", grp_id)
                return status

            app.groups[grp_id].remove_member(self)
            return status

    signature = {
        'endpoints': {
            1: {
                'profile_id': 0x0104,
                'device_type': 0xbeef,
                'input_clusters': [],
                'output_clusters': [zigpy.zcl.clusters.security.IasZone.cluster_id]
            },
        },
    }

    replacement = {
        'manufacturer': 'Silicon Labs',
        'model': 'EZSP',
        'endpoints': {
            1: (EZSPEndpoint, {})
        }
    }



import os

import click

from . import util

CHANNELS = list(range(11, 27))

channel = click.option(
    '-c', '--channel',
    type=click.IntRange(11, 26),
    required=True,
    default=15,
)

channels = click.option(
    '-C', '--channels',
    type=util.CSVParamType(11, 26),
    metavar='CHANNELS',
    default=','.join(map(str, CHANNELS)),
    show_default=True,
)

device = click.option(
    '-d', '--device',
    type=click.STRING,
    envvar='EZSP_DEVICE',
    required=True,
)

baudrate = click.option(
    '-b', '--baudrate',
    type=click.INT,
    envvar='EZSP_BAUDRATE',
    default=57600
)

database_file = click.option(
    '-D', '--database',
    type=click.Path(
        exists=True,
        dir_okay=False,
        writable=True,
    ),
    required=True,
    default=os.path.join(click.get_app_dir("bellows"), "app.db"),
)

duration_ms = click.option(
    '-t', '--duration', 'duration_ms',
    type=click.INT, metavar='MILLISECONDS',
    default=50,
    show_default=True,
)

duration_s = click.option(
    '-t', '--duration', 'duration_s',
    type=click.INT, metavar='SECONDS',
    default=30,
    show_default=True,
)

extended_pan = click.option(
    '-E', '--extended-pan-id',
    type=click.STRING,
)

pan = click.option(
    '-P', '--pan-id',
    type=click.IntRange(0, 65535),
)

manufacturer = click.option(
    '-m', '--manufacturer',
    type=click.IntRange(0, 0xFFFF),
    default=None,
    help='send a manufacturer specific command'
)



import bellows.types as t

COMMANDS = {
    # 4. Configuration frames
    'version': (0x00, (t.uint8_t, ), (t.uint8_t, t.uint8_t, t.uint16_t)),
    'getConfigurationValue': (0x52, (t.EzspConfigId, ), (t.EzspStatus, t.uint16_t)),
    'setConfigurationValue': (0x53, (t.EzspConfigId, t.uint16_t), (t.EzspStatus, )),
    'addEndpoint': (0x02, (t.uint8_t, t.uint16_t, t.uint16_t, t.uint8_t,
                           t.uint8_t, t.uint8_t, t.List(t.uint16_t,),
                           t.List(t.uint16_t)), (t.EzspStatus, )),
    'setPolicy': (0x55, (t.EzspPolicyId, t.EzspDecisionId), (t.EzspStatus, )),
    'getPolicy': (0x56, (t.EzspPolicyId, ), (t.EzspStatus, t.EzspDecisionId)),
    'getValue': (0xAA, (t.EzspValueId, ), (t.EzspStatus, t.LVBytes)),
    'getExtendedValue': (0x03, (t.EzspExtendedValueId, t.uint32_t), (t.EzspStatus, t.LVBytes)),
    'setValue': (0xAB, (t.EzspValueId, t.LVBytes), (t.EzspStatus, )),
    'setGpioCurrentConfiguration': (0xAC, (t.uint8_t, t.uint8_t, t.uint8_t), (t.EzspStatus, )),
    'setGpioPowerUpDownConfiguration': (0xAD, (t.uint8_t, t.uint8_t, t.uint8_t, t.uint8_t, t.uint8_t), (t.EzspStatus, )),
    'setGpioRadioPowerMask': (0xAE, (t.uint32_t, ), ()),
    'setCtune': (0xF5, (t.uint16_t, ), ()),
    'getCtune': (0xF6, (), (t.uint16_t, )),
    'setChannelMap': (0xF7, (t.uint8_t, t.uint8_t), ()),
    # 5. Utilities Frames
    'nop': (0x05, (), ()),
    'echo': (0x81, (t.LVBytes, ), (t.LVBytes, )),
    'invalidCommand': (0x58, (), (t.EzspStatus, )),
    'callback': (0x06, (), ()),
    'noCallbacks': (0x07, (), ()),
    'setToken': (0x09, (t.uint8_t, t.fixed_list(8, t.uint8_t)), (t.EmberStatus, )),
    'getToken': (0x0A, (t.uint8_t, ), (t.EmberStatus, t.fixed_list(8, t.uint8_t))),
    'getMfgToken': (0x0B, (t.EzspMfgTokenId, ), (t.LVBytes, )),
    'setMfgToken': (0x0C, (t.EzspMfgTokenId, t.LVBytes), (t.EmberStatus, )),
    'stackTokenChangedHandler': (0x0D, (), (t.uint16_t, )),
    'getRandomNumber': (0x49, (), (t.EmberStatus, t.uint16_t)),
    'setTimer': (0x0E, (t.uint8_t, t.uint16_t, t.EmberEventUnits, t.Bool), (t.EmberStatus, )),
    'getTimer': (0x4E, (t.uint8_t, ), (t.uint16_t, t.EmberEventUnits, t.Bool)),
    'timerHandler': (0x0F, (), (t.uint8_t, )),
    'debugWrite': (0x12, (t.Bool, t.LVBytes), (t.EmberStatus, )),
    'readAndClearCounters': (0x65, (), (t.fixed_list(t.EmberCounterType.COUNTER_TYPE_COUNT, t.uint16_t), )),
    'readCounters': (0xF1, (), (t.fixed_list(t.EmberCounterType.COUNTER_TYPE_COUNT, t.uint16_t), )),
    'counterRolloverHandler': (0xF2, (), (t.EmberCounterType, )),
    'delayTest': (0x9D, (t.uint16_t, ), ()),
    'getLibraryStatus': (0x01, (t.uint8_t, ), (t.EmberLibraryStatus, )),
    'getXncpInfo': (0x13, (), (t.EmberStatus, t.uint16_t, t.uint16_t)),
    'customFrame': (0x47, (t.LVBytes, ), (t.EmberStatus, t.LVBytes)),
    'customFrameHandler': (0x54, (), (t.LVBytes, )),
    'getEui64': (0x26, (), (t.EmberEUI64, )),
    'getNodeId': (0x27, (), (t.EmberNodeId, )),
    'networkInit': (0x17, (), (t.EmberStatus, )),
    # 6. Networking Frames
    'setManufacturerCode': (0x15, (t.uint16_t, ), ()),
    'setPowerDescriptor': (0x16, (t.uint16_t, ), ()),
    'networkInitExtended': (0x70, (t.EmberNetworkInitStruct, ), (t.EmberStatus, )),
    'networkState': (0x18, (), (t.EmberNetworkStatus, )),
    'stackStatusHandler': (0x19, (), (t.EmberStatus, )),
    'startScan': (0x1A, (t.EzspNetworkScanType, t.uint32_t, t.uint8_t), (t.EmberStatus, )),
    'energyScanResultHandler': (0x48, (), (t.uint8_t, t.int8s)),
    'networkFoundHandler': (0x1B, (), (t.EmberZigbeeNetwork, t.uint8_t, t.int8s)),
    'scanCompleteHandler': (0x1C, (), (t.uint8_t, t.EmberStatus)),
    'stopScan': (0x1D, (), (t.EmberStatus, )),
    'formNetwork': (0x1E, (t.EmberNetworkParameters, ), (t.EmberStatus, )),
    'joinNetwork': (0x1F, (t.EmberNodeType, t.EmberNetworkParameters), (t.EmberStatus, )),
    'leaveNetwork': (0x20, (), (t.EmberStatus, )),
    'findAndRejoinNetwork': (0x21, (t.Bool, t.uint32_t), (t.EmberStatus, )),
    'permitJoining': (0x22, (t.uint8_t, ), (t.EmberStatus, )),
    'childJoinHandler': (0x23, (), (t.uint8_t, t.Bool, t.EmberNodeId, t.EmberEUI64, t.EmberNodeType)),
    'energyScanRequest': (0x9C, (t.EmberNodeId, t.uint32_t, t.uint8_t, t.uint16_t), (t.EmberStatus, )),
    'getNetworkParameters': (0x28, (), (t.EmberStatus, t.EmberNodeType, t.EmberNetworkParameters)),
    'getParentChildParameters': (0x29, (), (t.uint8_t, t.EmberEUI64, t.EmberNodeId)),
    'getChildData': (0x4A, (t.uint8_t, ), (t.EmberStatus, t.EmberNodeId, t.EmberEUI64, t.EmberNodeType)),
    'getNeighbor': (0x79, (t.uint8_t, ), (t.EmberStatus, t.EmberNeighborTableEntry)),
    'neighborCount': (0x7A, (), (t.uint8_t, )),
    'getRouteTableEntry': (0x7B, (t.uint8_t, ), (t.EmberStatus, t.EmberRouteTableEntry)),
    'setRadioPower': (0x99, (t.int8s, ), (t.EmberStatus, )),
    'setRadioChannel': (0x9A, (t.uint8_t, ), (t.EmberStatus, )),
    'setConcentrator': (0x10, (t.Bool, t.uint16_t, t.uint16_t, t.uint16_t, t.uint8_t, t.uint8_t, t.uint8_t), (t.EmberStatus, )),
    # 7. Binding Frames
    'clearBindingTable': (0x2A, (), (t.EmberStatus, )),
    'setBinding': (0x2B, (t.uint8_t, t.EmberBindingTableEntry), (t.EmberStatus, )),
    'getBinding': (0x2C, (t.uint8_t, ), (t.EmberStatus, t.EmberBindingTableEntry)),
    'deleteBinding': (0x2D, (t.uint8_t, ), (t.EmberStatus, )),
    'bindingIsActive': (0x2E, (t.uint8_t, ), (t.Bool, )),
    'getBindingRemoteNodeId': (0x2F, (t.uint8_t, ), (t.EmberNodeId, )),
    'setBindingRemoteNodeId': (0x30, (t.uint8_t, ), ()),
    'remoteSetBindingHandler': (0x31, (), (t.EmberBindingTableEntry, )),
    'remoteDeleteBindingHandler': (0x32, (), (t.uint8_t, t.EmberStatus)),
    # 8. Messaging Frames
    'maximumPayloadLength': (0x33, (), (t.uint8_t, )),
    'sendUnicast': (0x34, (t.EmberOutgoingMessageType, t.EmberNodeId, t.EmberApsFrame, t.uint8_t, t.LVBytes), (t.EmberStatus, t.uint8_t)),
    'sendBroadcast': (0x36, (t.EmberNodeId, t.EmberApsFrame, t.uint8_t, t.uint8_t, t.LVBytes), (t.EmberStatus, t.uint8_t)),
    'proxyBroadcast': (0x37, (t.EmberNodeId, t.EmberNodeId, t.uint8_t, t.EmberApsFrame, t.uint8_t, t.uint8_t, t.LVBytes), (t.EmberStatus, t.uint8_t)),
    'sendMulticast': (0x38, (t.EmberApsFrame, t.uint8_t, t.uint8_t, t.uint8_t, t.LVBytes), (t.EmberStatus, t.uint8_t)),
    'sendReply': (0x39, (t.EmberNodeId, t.EmberApsFrame, t.LVBytes), (t.EmberStatus, )),
    'messageSentHandler': (0x3F, (), (t.EmberOutgoingMessageType, t.uint16_t, t.EmberApsFrame, t.uint8_t, t.EmberStatus, t.LVBytes)),
    'sendManyToOneRouteRequest': (0x41, (t.uint16_t, t.uint8_t), (t.EmberStatus, )),
    'pollForData': (0x42, (t.uint16_t, t.EmberEventUnits, t.uint8_t), (t.EmberStatus, )),
    'pollCompleteHandler': (0x43, (), (t.EmberStatus, )),
    'pollHandler': (0x44, (), (t.EmberNodeId, )),
    'incomingSenderEui64Handler': (0x62, (), (t.EmberEUI64, )),
    'incomingMessageHandler': (0x45, (), (t.EmberIncomingMessageType, t.EmberApsFrame, t.uint8_t, t.int8s, t.EmberNodeId, t.uint8_t, t.uint8_t, t.LVBytes)),
    'incomingRouteRecordHandler': (0x59, (), (t.EmberNodeId, t.EmberEUI64, t.uint8_t, t.int8s, t.LVList(t.EmberNodeId))),
    'changeSourceRouteHandler': (0xC4, (), (t.EmberNodeId, t.EmberNodeId, t.Bool)),
    'setSourceRoute': (0x5A, (t.EmberNodeId, t.LVList(t.EmberNodeId)), (t.EmberStatus, )),
    'incomingManyToOneRouteRequestHandler': (0x7D, (), (t.EmberNodeId, t.EmberEUI64, t.uint8_t)),
    'incomingRouteErrorHandler': (0x80, (), (t.EmberStatus, t.EmberNodeId)),
    'addressTableEntryIsActive': (0x5B, (t.uint8_t, ), (t.Bool, )),
    'setAddressTableRemoteEui64': (0x5C, (t.uint8_t, t.EmberEUI64), (t.EmberStatus, )),
    'setAddressTableRemoteNodeId': (0x5D, (t.uint8_t, t.EmberNodeId), ()),
    'getAddressTableRemoteEui64': (0x5E, (t.uint8_t, ), (t.EmberEUI64, )),
    'getAddressTableRemoteNodeId': (0x5F, (t.uint8_t, ), (t.EmberNodeId, )),
    'setExtendedTimeout': (0x7E, (t.EmberEUI64, t.Bool), ()),
    'getExtendedTimeout': (0x7F, (t.EmberEUI64, ), (t.Bool, )),
    'replaceAddressTableEntry': (0x82, (t.uint8_t, t.EmberEUI64, t.EmberNodeId, t.Bool, ), (t.EmberStatus, t.EmberEUI64, t.EmberNodeId, t.Bool, )),
    'lookupNodeIdByEui64': (0x60, (t.EmberEUI64, ), (t.EmberNodeId, )),
    'lookupEui64ByNodeId': (0x61, (t.EmberNodeId, ), (t.EmberStatus, t.EmberEUI64)),
    'getMulticastTableEntry': (0x63, (t.uint8_t, ), (t.EmberStatus, t.EmberMulticastTableEntry)),
    'setMulticastTableEntry': (0x64, (t.uint8_t, t.EmberMulticastTableEntry), (t.EmberStatus, )),
    'idConflictHandler': (0x7C, (), (t.EmberNodeId, )),
    'sendRawMessage': (0x96, (t.LVBytes, ), (t.EmberStatus, )),
    'macPassthroughMessageHandler': (0x97, (), (t.EmberMacPassthroughType, t.uint8_t, t.int8s, t.LVBytes)),
    'macFilterMatchMessageHandler': (0x46, (), (t.uint8_t, t.EmberMacPassthroughType, t.uint8_t, t.int8s, t.LVBytes)),
    'rawTransmitCompleteHandler': (0x98, (), (t.EmberStatus, )),
    # 9. Security Frames
    'setInitialSecurityState': (0x68, (t.EmberInitialSecurityState, ), (t.EmberStatus, )),
    'getCurrentSecurityState': (0x69, (), (t.EmberStatus, t.EmberCurrentSecurityState)),
    'getKey': (0x6A, (t.EmberKeyType, ), (t.EmberStatus, t.EmberKeyStruct)),
    'switchNetworkKeyHandler': (0x6E, (), (t.uint8_t, )),
    'getKeyTableEntry': (0x71, (t.uint8_t, ), (t.EmberStatus, t.EmberKeyStruct)),
    'setKeyTableEntry': (0x72, (t.uint8_t, t.EmberEUI64, t.Bool, t.EmberKeyData), (t.EmberStatus, )),
    'findKeyTableEntry': (0x75, (t.EmberEUI64, t.Bool), (t.uint8_t, )),
    'addOrUpdateKeyTableEntry': (0x66, (t.EmberEUI64, t.Bool, t.EmberKeyData), (t.EmberStatus, )),
    'eraseKeyTableEntry': (0x76, (t.uint8_t, ), (t.EmberStatus, )),
    'clearKeyTable': (0xB1, (), (t.EmberStatus, )),
    'requestLinkKey': (0x14, (t.EmberEUI64, ), (t.EmberStatus, )),
    'zigbeeKeyEstablishmentHandler': (0x9B, (), (t.EmberEUI64, t.EmberKeyStatus)),
    'addTransientLinkKey': (0xAF, (t.EmberEUI64, t.EmberKeyData), (t.EmberStatus, )),
    'clearTransientLinkKeys': (0x6B, (), ()),
    'getTransientLinkKey': (0xCE, (t.EmberEUI64, ), (t.EmberStatus, t.EmberTransientKeyData)),
    'setSecurityKey': (0xCA, (t.EmberKeyData, t.SecureEzspSecurityType), (t.EzspStatus, )),
    'setSecurityParameters': (0xCB, (t.SecureEzspSecurityLevel, t.SecureEzspRandomNumber), (t.EzspStatus, t.SecureEzspRandomNumber)),
    'resetToFactoryDefaults': (0xCC, (), (t.EzspStatus, )),
    'getSecurityKeyStatus': (0xCD, (), (t.EzspStatus, t.SecureEzspSecurityType)),
    # 10. Trust Center Frames
    'trustCenterJoinHandler': (0x24, (), (t.EmberNodeId, t.EmberEUI64, t.EmberDeviceUpdate, t.EmberJoinDecision, t.EmberNodeId)),
    'broadcastNextNetworkKey': (0x73, (t.EmberKeyData, ), (t.EmberStatus, )),
    'broadcastNetworkKeySwitch': (0x74, (), (t.EmberStatus, )),
    'becomeTrustCenter': (0x77, (t.EmberKeyData, ), (t.EmberStatus, )),
    'aesMmoHash': (0x6F, (t.EmberAesMmoHashContext, t.Bool, t.LVBytes), (t.EmberStatus, t.EmberAesMmoHashContext)),
    'removeDevice': (0xA8, (t.EmberNodeId, t.EmberEUI64, t.EmberEUI64), (t.EmberStatus, )),
    'unicastNwkKeyUpdate': (0xA9, (t.EmberNodeId, t.EmberEUI64, t.EmberKeyData), (t.EmberStatus, )),
    # 11. Certificate Based Key Exchange (CBKE) Frames
    'generateCbkeKeys': (0xA4, (), (t.EmberStatus, )),
    'generateCbkeKeysHandler': (0x9E, (), (t.EmberStatus, t.EmberPublicKeyData)),
    'calculateSmacs': (0x9F, (t.Bool, t.EmberCertificateData, t.EmberPublicKeyData), (t.EmberStatus, )),
    'calculateSmacsHandler': (0xA0, (), (t.EmberStatus, t.EmberSmacData, t.EmberSmacData)),
    'generateCbkeKeys283k1': (0xE8, (), (t.EmberStatus, )),
    'generateCbkeKeysHandler283k1': (0xE9, (), (t.EmberStatus, t.EmberPublicKey283k1Data)),
    'calculateSmacs283k1': (0xEA, (t.Bool, t.EmberCertificate283k1Data, t.EmberPublicKey283k1Data), (t.EmberStatus, )),
    'calculateSmacsHandler283k1': (0xEB, (), (t.EmberStatus, t.EmberSmacData, t.EmberSmacData)),
    'clearTemporaryDataMaybeStoreLinkKey': (0xA1, (t.Bool, ), (t.EmberStatus, )),
    'clearTemporaryDataMaybeStoreLinkKey283k1': (0xEE, (t.Bool, ), (t.EmberStatus, )),
    'getCertificate': (0xA5, (), (t.EmberStatus, t.EmberCertificateData)),
    'getCertificate283k1': (0xEC, (), (t.EmberStatus, t.EmberCertificate283k1Data)),
    'dsaSign': (0xA6, (t.LVBytes, ), (t.EmberStatus, )),  # Deprecated
    'dsaSignHandler': (0xA7, (), (t.EmberStatus, t.LVBytes)),  # Deprecated
    'dsaVerify': (0xA3, (t.EmberMessageDigest, t.EmberCertificateData, t.EmberSignatureData), (t.EmberStatus, )),
    'dsaVerifyHandler': (0x78, (), (t.EmberStatus, )),
    'dsaVerify283k1': (0xB0, (t.EmberMessageDigest, t.EmberCertificate283k1Data, t.EmberSignature283k1Data), (t.EmberStatus, )),
    'setPreinstalledCbkeData': (0xA2, (t.EmberPublicKeyData, t.EmberCertificateData, t.EmberPrivateKeyData), (t.EmberStatus, )),
    'setPreinstalledCbkeData283k1': (0xED, (t.EmberPublicKey283k1Data, t.EmberCertificate283k1Data, t.EmberPrivateKey283k1Data), (t.EmberStatus, )),
    # 12. Mfglib Frames
    'mfglibStart': (0x83, (t.Bool, ), (t.EmberStatus, )),
    'mfglibEnd': (0x84, (), (t.EmberStatus, )),
    'mfglibStartTone': (0x85, (), (t.EmberStatus, )),
    'mfglibStopTone': (0x86, (), (t.EmberStatus, )),
    'mfglibStartStream': (0x87, (), (t.EmberStatus, )),
    'mfglibStopStream': (0x88, (), (t.EmberStatus, )),
    'mfglibSendPacket': (0x89, (t.LVBytes, ), (t.EmberStatus, )),
    'mfglibSetChannel': (0x8A, (t.uint8_t, ), (t.EmberStatus, )),
    'mfglibGetChannel': (0x8B, (), (t.uint8_t, )),
    'mfglibSetPower': (0x8C, (t.uint16_t, t.int8s), (t.EmberStatus, )),
    'mfglibGetPower': (0x8D, (), (t.int8s, )),
    'mfglibRxHandler': (0x8E, (), (t.uint8_t, t.int8s, t.LVBytes)),
    # 13. Bootloader Frames
    'launchStandaloneBootloader': (0x8F, (t.uint8_t, ), (t.EmberStatus, )),
    'sendBootloadMessage': (0x90, (t.Bool, t.EmberEUI64, t.LVBytes), (t.EmberStatus, )),
    'getStandaloneBootloaderVersionPlatMicroPhy': (0x91, (), (t.uint16_t, t.uint8_t, t.uint8_t, t.uint8_t)),
    'incomingBootloadMessageHandler': (0x92, (), (t.EmberEUI64, t.uint8_t, t.int8s, t.LVBytes)),
    'bootloadTransmitCompleteHandler': (0x93, (), (t.EmberStatus, t.LVBytes)),
    'aesEncrypt': (0x94, (t.fixed_list(16, t.uint8_t), t.fixed_list(16, t.uint8_t)), (t.fixed_list(16, t.uint8_t), )),
    'overrideCurrentChannel': (0x95, (t.uint8_t, ), (t.EmberStatus, )),
    # 14. ZLL Frames
    'zllNetworkOps': (0xB2, (t.EmberZllNetwork, t.EzspZllNetworkOperation, t.int8s), (t.EmberStatus, )),
    'zllSetInitialSecurityState': (0xB3, (t.EmberKeyData, t.EmberZllInitialSecurityState), (t.EmberStatus, )),
    'zllStartScan': (0xB4, (t.uint32_t, t.int8s, t.EmberNodeType), (t.EmberStatus, )),
    'zllSetRxOnWhenIdle': (0xB5, (t.uint16_t, ), (t.EmberStatus, )),
    'zllNetworkFoundHandler': (0xB6, (), (t.EmberZllNetwork, t.Bool, t.EmberZllDeviceInfoRecord, t.uint8_t, t.int8s)),
    'zllScanCompleteHandler': (0xB7, (), (t.EmberStatus, )),
    'zllAddressAssignmentHandler': (0xB8, (), (t.EmberZllAddressAssignment, t.uint8_t, t.int8s)),
    'setLogicalAndRadioChannel': (0xB9, (t.uint8_t, ), (t.EmberStatus, )),
    'getLogicalChannel': (0xBA, (), (t.uint8_t, )),
    'zllTouchLinkTargetHandler': (0xBB, (), (t.EmberZllNetwork, )),
    'zllGetTokens': (0xBC, (), (t.EmberTokTypeStackZllData, t.EmberTokTypeStackZllSecurity)),
    'zllSetDataToken': (0xBD, (t.EmberTokTypeStackZllData, ), ()),
    'zllSetNonZllNetwork': (0xBF, (), ()),
    'isZllNetwork': (0xBE, (), (t.Bool, )),
    # 15 RF4CE Frames
    'rf4ceSetPairingTableEntry': (0xD0, (t.uint8_t, t.EmberRf4cePairingTableEntry), (t.EmberStatus, )),
    'rf4ceGetPairingTableEntry': (0xD1, (t.uint8_t, ), (t.EmberStatus, t.EmberRf4cePairingTableEntry)),
    'rf4ceDeletePairingTableEntry': (0xD2, (t.uint8_t, ), (t.EmberStatus, )),
    'rf4ceKeyUpdate': (0xD3, (t.uint8_t, t.EmberKeyData), (t.EmberStatus, )),
    'rf4ceSend': (0xD4, (t.uint8_t, t.uint8_t, t.uint16_t, t.EmberRf4ceTxOption, t.uint8_t, t.LVBytes), (t.EmberStatus, )),
    'rf4ceIncomingMessageHandler': (0xD5, (), (t.uint8_t, t.uint8_t, t.uint16_t, t.EmberRf4ceTxOption, t.LVBytes)),
    'rf4ceMessageSentHandler': (0xD6, (), (t.EmberStatus, t.uint8_t, t.EmberRf4ceTxOption, t.uint8_t, t.uint16_t, t.uint8_t, t.LVBytes)),
    'rf4ceStart': (0xD7, (t.EmberRf4ceNodeCapabilities, t.EmberRf4ceVendorInfo, t.int8s), (t.EmberStatus, )),
    'rf4ceStop': (0xD8, (), (t.EmberStatus, )),
    'rf4ceDiscovery': (0xD9, (t.EmberPanId, t.EmberNodeId, t.uint8_t, t.uint16_t, t.LVBytes), (t.EmberStatus, )),
    'rf4ceDiscoveryCompleteHandler': (0xDA, (), (t.EmberStatus, )),
    'rf4ceDiscoveryRequestHandler': (0xDB, (), (t.EmberEUI64, t.uint8_t, t.EmberRf4ceVendorInfo, t.EmberRf4ceApplicationInfo, t.uint8_t, t.uint8_t)),
    'rf4ceDiscoveryResponseHandler': (0xDC, (), (t.Bool, t.uint8_t, t.EmberPanId, t.EmberEUI64, t.uint8_t, t.EmberRf4ceVendorInfo, t.EmberRf4ceApplicationInfo, t.uint8_t, t.uint8_t)),
    'rf4ceEnableAutoDiscoveryResponse': (0xDD, (t.uint16_t, ), (t.EmberStatus, )),
    'rf4ceAutoDiscoveryResponseCompleteHandler': (0xDE, (), (t.EmberStatus, t.EmberEUI64, t.uint8_t, t.EmberRf4ceVendorInfo, t.EmberRf4ceApplicationInfo, t.uint8_t)),
    'rf4cePair': (0xDF, (t.uint8_t, t.EmberPanId, t.EmberEUI64, t.uint8_t), (t.EmberStatus, )),
    'rf4cePairCompleteHandler': (0xE0, (), (t.EmberStatus, t.uint8_t, t.EmberRf4ceVendorInfo, t.EmberRf4ceApplicationInfo)),
    'rf4cePairRequestHandler': (0xE1, (), (t.EmberStatus, t.uint8_t, t.EmberEUI64, t.uint8_t, t.EmberRf4ceVendorInfo, t.EmberRf4ceApplicationInfo, t.uint8_t)),
    'rf4ceUnpair': (0xE2, (t.uint8_t, ), (t.EmberStatus, )),
    'rf4ceUnpairHandler': (0xE3, (), (t.uint8_t, )),
    'rf4ceUnpairCompleteHandler': (0xE4, (), (t.uint8_t, )),
    'rf4ceSetPowerSavingParameters': (0xE5, (t.uint32_t, t.uint32_t), (t.EmberStatus, )),
    'rf4ceSetFrequencyAgilityParameters': (0xE6, (t.uint8_t, t.uint8_t, t.int8s, t.uint16_t, t.uint8_t), (t.EmberStatus, )),
    'rf4ceSetApplicationInfo': (0xE7, (t.EmberRf4ceApplicationInfo, ), (t.EmberStatus, )),
    'rf4ceGetApplicationInfo': (0xEF, (), (t.EmberStatus, t.EmberRf4ceApplicationInfo)),
    'rf4ceGetMaxPayload': (0xF3, (t.uint8_t, t.EmberRf4ceTxOption), (t.uint8_t, )),
    'rf4ceGetNetworkParameters': (0xF4, (), (t.EmberStatus, t.EmberNodeType, t.EmberNetworkParameters)),
    # 16 Green Power Frames
    'gpProxyTableProcessGpPairing': (0xC9, (t.uint32_t, t.EmberGpAddress, t.uint8_t, t.uint16_t, t.uint16_t, t.uint16_t, t.fixed_list(8, t.uint8_t), t.EmberKeyData), ()),
    'dGpSend': (0xC6, (t.Bool, t.Bool, t.EmberGpAddress, t.uint8_t, t.LVBytes, t.uint8_t, t.uint16_t), (t.EmberStatus, )),
    'dGpSentHandler': (0xC7, (), (t.EmberStatus, t.uint8_t)),
    'gpepIncomingMessageHandler': (0xC5, (), (t.EmberStatus, t.uint8_t, t.uint8_t, t.EmberGpAddress, t.EmberGpSecurityLevel, t.EmberGpKeyType, t.Bool, t.Bool, t.uint32_t, t.uint8_t, t.uint32_t, t.EmberGpSinkListEntry, t.LVBytes)),
}



"""Setup module for bellows"""

from setuptools import find_packages, setup

import bellows

setup(
    name="bellows-homeassistant",
    version=bellows.__version__,
    description="Library implementing EZSP",
    url="http://github.com/zigpy/bellows",
    author="Russell Cloran",
    author_email="rcloran@gmail.com",
    license="GPL-3.0",
    packages=find_packages(exclude=['*.tests']),
    entry_points={
        'console_scripts': ['bellows=bellows.cli.main:main'],
    },
    install_requires=[
        'Click',
        'click-log==0.2.0',
        'pure_pcapy3==1.0.1',
        'pyserial-asyncio',
        'zigpy-homeassistant>=0.6.0',
    ],
    dependency_links=[
        'https://github.com/rcloran/pure-pcapy-3/archive/e289c7d7566306dc02d8f4eb30c0358b41f40f26.zip#egg=pure_pcapy3-1.0.1',
    ],
    tests_require=[
        'pytest',
    ],
)
